from setuptools import setup, find_packages

setup(
    name="cluster_algebra",
    version="0.1.1",
    packages=find_packages(),
    install_requires=[
        "numpy>=2.0.2",
        "pandas>=2.2.3",
        "matplotlib>=3.9.2",
        "networkx>=3.2.1",
        "scipy>=1.13.1",
        "sympy>=1.13.3",
        "shapely>=2.0.6",
        "pydantic>=2.9.2",
    ],
    extras_require={
        'dev': [
            'pytest>=8.3.3',
        ],
    },
    description="calculator for cluster algebras",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    python_requires='>=3.9',
)